package com.loopperfect.buckaroo;

import com.google.common.util.concurrent.SettableFuture;
import com.loopperfect.buckaroo.cli.CLICommand;
import com.loopperfect.buckaroo.cli.CLIParsers;
import com.loopperfect.buckaroo.tasks.LoggingTasks;
import com.loopperfect.buckaroo.views.ProgressView;
import com.loopperfect.buckaroo.views.StatsView;
import com.loopperfect.buckaroo.views.SummaryView;
import com.loopperfect.buckaroo.virtualterminal.TerminalBuffer;
import com.loopperfect.buckaroo.virtualterminal.components.Component;
import com.loopperfect.buckaroo.virtualterminal.components.StackLayout;
import io.reactivex.Observable;
import io.reactivex.schedulers.Schedulers;
import org.fusesource.jansi.AnsiConsole;
import org.jparsec.Parser;

import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.io.PrintStream;

import static com.loopperfect.buckaroo.ErrorHandler.handleErrors;

public final class Main {

    public static final int TERMINAL_WIDTH = 80;

    private Main() {}

    public static void main(final String[] args) {

        PrintStream p = System.out;

        if (args.length == 0) {
            p.println("   ___           __                             ");
            p.println("  / _ )__ ______/ /_____ ________  ___          ");
            p.println(" / _  / // / __/  '_/ _ `/ __/ _ \\/ _ \\       ");
            p.println("/____/\\_,_/\\__/_/\\_\\\\_,_/_/  \\___/\\___/\n");

            p.println("Buck, Buck, Buckaroo! \uD83E\uDD20");
            p.println("https://buckaroo.readthedocs.io/");
            return;
        }

        final FileSystem fs = FileSystems.getDefault();

        final String rawCommand = String.join(" ", args);

        final CountDownLatch loggingLatch = new CountDownLatch(1);
        final CountDownLatch taskLatch = new CountDownLatch(1);

        // Send the command to the logging server, if present
        try {

            LoggingTasks.log(fs, rawCommand).subscribe(
                next -> {
                    // Do nothing
                },
                error -> {
                    // Do nothing
                    loggingLatch.countDown();
                },
                () -> {
                    // Do nothing
                    loggingLatch.countDown();
                });

        } catch (final Throwable ignored) {
          loggingLatch.countDown();
        }

        // Parse the command
        final Parser<CLICommand> commandParser = CLIParsers.commandParser;

        try {
            final CLICommand command = commandParser.parse(rawCommand);

            final Observable<Event> task = command.routine().apply(fs);

            // Our renderers can't handle errors properly, so let's either-ize the stream
            // and split it to two stream in publish.
            // TODO: Fix error-handling in the renders
            final Observable<Either<Throwable, Event>> errorOrEvent = task
                .map(Either::<Throwable, Event>right)
                .onErrorReturn(Either::<Throwable, Event>left);

            final Observable<Component> components = errorOrEvent
                .publish(upstream -> {
                        // It's crucial that we don't subscribe multiple times to our event emitters as they do I/O.
                        Observable<Component> errors  = upstream
                            .filter(Either::isLeft)
                            .map(x -> x.left().get())
                            .cache() // upstream is a hot observable, we want make sure we don't lose the error
                            .flatMap(Observable::error)
                            .cast(Component.class);

                        return upstream
                            .takeUntil(errors) // stop when an error occurs, takeUntil also propagates the error further down the chain.
                            .filter(Either::isRight)
                            .map(e -> e.right().get())
                            .compose(u -> Observable.combineLatest(
                                    ProgressView.render(u)
                                        .startWith(StackLayout.of())
                                        .subscribeOn(Schedulers.computation())
                                        .concatWith(Observable.just(StackLayout.of())),
                                    StatsView.render(u)
                                        .subscribeOn(Schedulers.computation())
                                        .skip(300, TimeUnit.MILLISECONDS)
                                        .takeUntil(upstream.lastElement().toObservable())
                                        .startWith(StackLayout.of()),
                                    SummaryView.render(u)
                                        .takeLast(1)
                                        .startWith(StackLayout.of()),
                                (x, y, z) -> (Component) StackLayout.of(x, y, z))
                        );
                })
                .subscribeOn(Schedulers.computation())
                .sample(100, TimeUnit.MILLISECONDS, true) // Most terminals can't handle more than 10fps
                .distinctUntilChanged();

            AnsiConsole.systemInstall();

            final TerminalBuffer buffer = new TerminalBuffer();

            final SettableFuture<Throwable> errorF = SettableFuture.create();

            components
                .map(x -> x.render(TERMINAL_WIDTH))
                .subscribe(
                    str -> {
                        buffer.flip(str);
                        // This is only a hint for garbage collection.
                        // This is an opportune moment for collection as we accumulated a lot of Components.
                        // The next render wont happen in the next 100ms anyway...
                        System.gc();
                    },
                    error -> {
                        errorF.set(error);
                        taskLatch.countDown();
                    },
                    () -> {
                        taskLatch.countDown();
                    });

            // Ensure we don't exit main before we have successfully unsubscribed from the stream
            taskLatch.await();

            if (errorF.isDone()) {
                // It turns out that RxJava does not guarantee that onNext won't be called after onError in a multi-threaded enviroment.
                // Let's store the error in a future, wait until all events have stopped firing and render a message to the user.
                handleErrors(errorF.get(), buffer, fs);
            }

            try {
                loggingLatch.await(1000L, TimeUnit.MILLISECONDS);
            } catch (final Throwable ignored) {

            }

        } catch (final Throwable e) {
            p.println("Uh oh!");
            p.println(e.getMessage());
            p.println("Visit https://github.com/LoopPerfect/buckaroo/issues for help");
        }
    }
}
